from .Node import Node


class Graph:
    """
    Класс, реализующий граф для использования в алгоритме Дейкстры, а также сам алгоритм

    Данное решение является заимствованным
    Источник: https://gist.github.com/micahshute/bc8b45020636d862105543ecb231b9d2#file-adjacency_matrix_graphy-py
    """

    @classmethod
    def create_from_nodes(cls, nodes):
        return Graph(len(nodes), len(nodes), nodes)

    def __init__(self, row, col, nodes=None):
        self.adj_mat = [[0] * col for _ in range(row)]
        self.nodes = nodes
        for i in range(len(self.nodes)):
            self.nodes[i].index = i

    def connect_dir(self, node1, node2, weight=1):
        node1, node2 = self.get_index_from_node(node1), self.get_index_from_node(node2)
        self.adj_mat[node1][node2] = weight

    def connect(self, node1, node2, weight=1):
        self.connect_dir(node1, node2, weight)

    def connections_from(self, node):
        node = self.get_index_from_node(node)
        return [(self.nodes[col_num], self.adj_mat[node][col_num]) for col_num in range(len(self.adj_mat[node])) if
                self.adj_mat[node][col_num] != 0]

    def connections_to(self, node):
        node = self.get_index_from_node(node)
        column = [row[node] for row in self.adj_mat]
        return [(self.nodes[row_num], column[row_num]) for row_num in range(len(column)) if column[row_num] != 0]

    def print_adj_mat(self):
        for row in self.adj_mat:
            print(row)

    def node(self, index):
        return self.nodes[index]

    def remove_conn(self, node1, node2):
        self.remove_conn_dir(node1, node2)
        self.remove_conn_dir(node2, node1)

    def remove_conn_dir(self, node1, node2):
        node1, node2 = self.get_index_from_node(node1), self.get_index_from_node(node2)
        self.adj_mat[node1][node2] = 0

    def can_traverse_dir(self, node1, node2):
        node1, node2 = self.get_index_from_node(node1), self.get_index_from_node(node2)
        return self.adj_mat[node1][node2] != 0

    def has_conn(self, node1, node2):
        return self.can_traverse_dir(node1, node2) or self.can_traverse_dir(node2, node1)

    def add_node(self, node):
        self.nodes.append(node)
        node.index = len(self.nodes) - 1
        for row in self.adj_mat:
            row.append(0)
        self.adj_mat.append([0] * (len(self.adj_mat) + 1))

    def get_weight(self, n1, n2):
        node1, node2 = self.get_index_from_node(n1), self.get_index_from_node(n2)
        return self.adj_mat[node1][node2]

    @staticmethod
    def get_index_from_node(node):
        if not isinstance(node, (Node, int)):
            raise ValueError("node must be an integer or a Node object")
        if isinstance(node, int):
            return node
        else:
            return node.index

    def dijkstra(self, node):
        nodenum = self.get_index_from_node(node)
        dist = [None] * len(self.nodes)

        for i in range(len(dist)):
            dist[i] = [float("inf")]
            dist[i].append([self.nodes[nodenum]])

        dist[nodenum][0] = 0
        queue, seen = [i for i in range(len(self.nodes))], set()

        while len(queue) > 0:
            min_dist, min_node = float("inf"), None

            for n in queue:
                if dist[n][0] < min_dist and n not in seen:
                    min_dist, min_node = dist[n][0], n

            if min_node is None:
                break

            queue.remove(min_node)
            seen.add(min_node)

            connections = self.connections_from(min_node)

            for (node, weight) in connections:
                tot_dist = weight + min_dist
                if tot_dist < dist[node.index][0]:
                    dist[node.index][0] = tot_dist
                    dist[node.index][1] = list(dist[min_node][1])
                    dist[node.index][1].append(node)

        return dist



"""
Модуль itmo-routespd позволяет строить индивидуальные образовательные маршруты
для помощи абитуриентам в выборе существующих
направлений подготовки, факультетов и рабочих программ дисциплин.

Vladislav Pivosh [https://github.com/RUGyron]
pivosh098@gmail.com
"""

__author__ = 'Vladislav Pivosh'

__version__ = '1.0'

from .BaseRouter import *
from .Data import *
from .DetailedRouter import *



from functools import reduce
from typing import Iterator

from .Data import Data
from .funcs import count_base_coefficient, count_prefer_coefficient, get_fields_of_studies4rpd, get_rpd_id


class BaseRouter:
    """
    Класс для построения общего маршрута по всем РПД.
    """

    def __init__(self, data: Data):
        super().__init__()

        self._data = data
        self._rpd_id_list = set(data.rpd_names.keys())

        self.__reset_params__()

    def __reset_params__(self):
        self.__learned_items = self._data.learned
        self.__learned_rpd = set()

    @property
    def learned_rpd_count(self) -> int:
        """
        :return: Количество изученных РПД
        """

        return self.__learned_rpd.__len__()

    @property
    def learned_items_count(self) -> int:
        """
        :return: Количество освоенных тем
        """

        return self.__learned_items.__len__()

    def produce(self, threshold: float, iter_count: int, rpd_count: int) -> Iterator:
        """
        Метод для построения общего маршрута по всем РПД для абитуриента.

        :param threshold: порог для коэффициента допущения РПД к рассмотрению. Чем больше, тем меньше РПД пройдёт
        :param iter_count: количество итераций, где каждая представляет собой набор РПД
        :param rpd_count: количество РПД на итерацию
        :return: упорядоченный генератор списка наборов РПД
        """

        self.__reset_params__()

        assert isinstance(threshold, float) and 0 < threshold < 1
        assert isinstance(iter_count, int) and iter_count > 0
        assert isinstance(rpd_count, int) and rpd_count > 0

        for iter_num in range(iter_count):
            selection = list(map(
                lambda x: (x, count_base_coefficient(x, self.__learned_items, self._data.prerequisites)),
                self._rpd_id_list - self.__learned_rpd,
            ))

            available = list(filter(lambda x: x[1] >= threshold, selection))

            selection = list(map(
                lambda x: (x[0], count_prefer_coefficient(x[0], self._data.preferred, self._data.postrequisites)),
                available,
            ))

            selection.sort(key=lambda x: x[1], reverse=True)

            most_preferred = selection[: rpd_count]

            if not selection:
                break

            new_learned = list(filter(
                lambda x: x not in self.__learned_items,
                reduce(
                    lambda x, y: x + y,
                    [self._data.postrequisites[preferred_rpd[0]] for preferred_rpd in most_preferred],
                ),
            ))

            self.__learned_items.extend(new_learned)

            for rpd_id, _ in most_preferred:
                self.__learned_rpd.add(rpd_id)

            rpd_batch = []
            for rpd_details in [(self._data.rpd_names[rpd[0]], rpd[1]) for rpd in most_preferred]:
                rpd_batch.append(dict(
                    rpd_id=get_rpd_id(rpd_details[0], self._data.rpd_names),
                    rpd_title=rpd_details[0],
                    preferred_coefficient=rpd_details[1],
                    fields_of_studies=get_fields_of_studies4rpd(rpd_details[0], self._data, is_id=False),
                ))

            yield dict(
                learned_items_count=self.learned_items_count,
                learned_rpd_count=self.learned_rpd_count,
                rpd_batch=rpd_batch,
            )



class Node:
    """
    Класс, реализующий вершину в графе для использования в алгоритме Дейкстры

    Данное решение является заимствованным
    Источник: https://gist.github.com/micahshute/bc8b45020636d862105543ecb231b9d2#file-adjacency_matrix_graphy-py
    """

    def __init__(self, data, indexloc=None):
        self.data = data
        self.index = indexloc



from functools import reduce
from typing import Iterator, List, Tuple, Optional, Dict, Any

from .Data import Data
from .dijkstra import Node, Graph
from .funcs import get_rpd_id, get_fields_of_studies4rpd


class DetailedRouter:
    """
    Класс для построения детального маршрута для заданной РПД для абитуриента.
    """

    def __init__(self, data: Data):
        super().__init__()

        self._data = data

        self.__reset_params__()

    def __reset_params__(self):
        self.__graph_data = []

    def _get_nodes(self) -> List:
        return list(set(map(
            lambda x: Node(x),
            filter(
                lambda x: isinstance(x, int),
                reduce(
                    lambda x, y: x + y,
                    self.__graph_data,
                ),
            ),
        )))

    def _get_graph(self, nodes: List[Node]) -> Graph:
        graph = Graph.create_from_nodes(nodes)
        for relation in self.__graph_data:
            graph.connect(
                self._get_node(relation[1], nodes),
                self._get_node(relation[0], nodes),
                1 - relation[2],
            )

        return graph

    @staticmethod
    def _get_routes_from(graph: Graph, start_node: Node) -> List[Tuple[float, List[Dict[str, Any]]]]:
        return [(
            weight,
            [n.data for n in nodes],
        ) for weight, nodes in graph.dijkstra(start_node) if len(nodes) > 1]

    @staticmethod
    def _get_node(node_id: int, nodes: List[Node]) -> Node:
        return list(filter(lambda x: x.data == node_id, nodes))[0]

    def get_short_routes(self, rpd_title: str, threshold: float, routes_count: int) -> Optional[Iterator]:
        """
        Метод для построения детального маршрута для заданной РПД для абитуриента.

        :param rpd_title: название РПД
        :param threshold: порог для коэффициента покрытия связи в графе между РПД. Чем больше порог,
            тем меньше связей будет доступно
        :param routes_count: ограничение по максимальному количеству маршрутов в выдаче
        :return: упорядоченный генератор списка маршрутов к РПД
        """

        assert isinstance(rpd_title, str) and rpd_title in self._data.rpd_names.values()
        assert isinstance(threshold, float) and 0 < threshold < 1
        assert isinstance(routes_count, int) and routes_count > 0

        self.__graph_data.extend([relation for relation in self._data.graph_data if relation[2] >= threshold])

        nodes = self._get_nodes()
        graph = self._get_graph(nodes)

        try:
            start_node = self._get_node(get_rpd_id(rpd_title, self._data.rpd_names), nodes)
        except IndexError:
            return

        routes = sorted(self._get_routes_from(graph, start_node), key=lambda x: x[0])[: routes_count]

        for route in routes:
            yield dict(
                count=len(route[1]),
                rate=round(route[0], 2),
                route=list(map(
                    lambda x: dict(
                        rpd_id=x,
                        rpd_title=self._data.rpd_names[x],
                        fields_of_studies=get_fields_of_studies4rpd(x, self._data),
                    ),
                    route[1][::-1]),
                ),
            )



"""
from json import dumps
import psycopg2

from itmo_routespd import Data, BaseRouter, DetailedRouter
from itmo_routespd.funcs import preferred_rpd_list, get_fields_of_studies4rpd


conn = psycopg2.connect(
    user='postgres',
    password='some_pass',
    host='127.0.0.1',
    port='5432',
    database='analytics_db',
)
cursor = conn.cursor()

rpd_table = 'public.workprogramsapp_workprogram'
items_table = 'public.dataprocessing_items'
ppk_table = 'public.workprogramsapp_prerequisitesofworkprogram'
rpk_table = 'public.workprogramsapp_outcomesofworkprogram'
rpd_in_field_of_study_table = 'public.workprogramsapp_workprograminfieldofstudy'
rpd_change_in_rpd_block_module_table = 'public.workprogramsapp_workprogramchangeindisciplineblockmodule'
rpd_block_module_table = 'public.workprogramsapp_disciplineblockmodule'
rpd_block_table = 'public.workprogramsapp_disciplineblock'
academic_plan_table = 'public.workprogramsapp_academicplan'
implementation_academic_plan_table = 'public.workprogramsapp_implementationacademicplan'
m2m_implementation_academic_plan2field_of_study_table = 'public.workprogramsapp_implementationacademicplan_field_of_study'
workprogramsapp_fieldofstudy_table = 'public.workprogramsapp_fieldofstudy'

cursor.execute(f'select id, title from {rpd_table};')
rpd_selection = cursor.fetchall()
cursor.execute(f'select id, name from {items_table};')
items_selection = cursor.fetchall()
cursor.execute(f'select item_id, workprogram_id from {rpk_table};')
postrequisites_selection = cursor.fetchall()
cursor.execute(f'select item_id, workprogram_id from {ppk_table};')
prerequisites_selection = cursor.fetchall()
cursor.execute(f'select work_program_id, work_program_change_in_discipline_block_module_id from {rpd_in_field_of_study_table};')
rpd_in_field_of_study_selection = cursor.fetchall()
cursor.execute(f'select id, discipline_block_module_id from {rpd_change_in_rpd_block_module_table};')
rpd_change_in_rpd_block_module_selection = cursor.fetchall()
cursor.execute(f'select id, descipline_block_id from {rpd_block_module_table};')
rpd_block_module_selection = cursor.fetchall()
cursor.execute(f'select id, academic_plan_id from {rpd_block_table};')
rpd_block_selection = cursor.fetchall()
cursor.execute(f'select id from {academic_plan_table};')
academic_plan_selection = cursor.fetchall()
cursor.execute(f'select academic_plan_id, id from {implementation_academic_plan_table};')
implementation_academic_plan_selection = cursor.fetchall()
cursor.execute(f'select implementationacademicplan_id, fieldofstudy_id from {m2m_implementation_academic_plan2field_of_study_table};')
m2m_implementation_academic_plan2field_of_study_selection = cursor.fetchall()
cursor.execute(f'select id, number, qualification, title, faculty from {workprogramsapp_fieldofstudy_table};')
workprogramsapp_fieldofstudy_selection = cursor.fetchall()

if conn:
    cursor.close()
    conn.close()

data = Data(
    learned=[969, 970, 1412, 2042, 2325, 3864, 8832, 12089, 15659, 15696, 15697, 1404, 16356, 1411, 9752, 435, 2369, 2356, 4534, 15632, 5465, 18650, 4573, 9539, 18510, 293, 1027, 7504, 3222, 18662, 18813, 18431, 2396, 18501, 18699, 120, 273, 9556, 2854, 115, 5844, 16214, 18465, 9694, 728, 10318, 18661, 4647, 18449, 992, 6191, 18697, 3109, 9659, 9285, 7246, 18487, 18890, 18655, 18432, 18433, 2910, 18, 18568, 9383, 4600, 327, 2390, 949, 409, 18893, 301, 6205, 18397, 3448],
    preferred=[2, 3, 15, 16, 19, 251, 253, 254, 255, 256, 257, 274, 275, 276, 277, 2096, 8909, 18861, 16638, 1003, 18875, 7538, 9556, 15313, 18592, 17654, 2280, 15340, 4569, 16538, 6560, 121, 337, 6567, 16288],
    rpd_selection=rpd_selection,
    items_selection=items_selection,
    prerequisites_selection=prerequisites_selection,
    postrequisites_selection=postrequisites_selection,
    rpd_in_field_of_study_selection=rpd_in_field_of_study_selection,
    rpd_change_in_rpd_block_module_selection=rpd_change_in_rpd_block_module_selection,
    rpd_block_module_selection=rpd_block_module_selection,
    rpd_block_selection=rpd_block_selection,
    academic_plan_selection=academic_plan_selection,
    implementation_academic_plan_selection=implementation_academic_plan_selection,
    m2m_implementation_academic_plan2field_of_study_selection=m2m_implementation_academic_plan2field_of_study_selection,
    workprogramsapp_fieldofstudy_selection=workprogramsapp_fieldofstudy_selection,
)

for batch in BaseRouter(data).produce(.2, 4, 10):
    print(dumps(batch, indent=4, ensure_ascii=False))

for route in DetailedRouter(data).get_short_routes('Основы технологически адекватного применения пищевых добавок', .01, 10):
    print(dumps(route, indent=4, ensure_ascii=False))

for rpd in preferred_rpd_list(data.preferred, data.postrequisites, data, 10):
    print(dumps(rpd, indent=4, ensure_ascii=False))

for fos in get_fields_of_studies4rpd(12438, data):
    print(dumps(fos, indent=4, ensure_ascii=False))
"""


from gc import collect
from pydantic import BaseModel, root_validator
from typing import Tuple, List, Optional

from .funcs import count_weight


class Data(BaseModel):
    """
    Датакласс, хранящий все данные и гиперпараметры для построения маршрутов
    """

    # Перменные для заполнения вручную
    learned: List[int]
    preferred: List[int]

    graph_data: Tuple[Tuple[int, int, float]]

    # select id, title from public.workprogramsapp_workprogram;
    rpd_selection: List[Tuple[int, str]]

    # select id, name from public.dataprocessing_items;
    items_selection: List[Tuple[int, str]]

    # select item_id, workprogram_id from public.workprogramsapp_prerequisitesofworkprogram;
    prerequisites_selection: List[Tuple[int, int]]

    # select item_id, workprogram_id from public.workprogramsapp_outcomesofworkprogram;
    postrequisites_selection: List[Tuple[int, int]]

    # select work_program_id, work_program_change_in_discipline_block_module_id
    # from public.workprogramsapp_workprograminfieldofstudy;
    rpd_in_field_of_study_selection: List[Tuple[int, int]]

    # select id, discipline_block_module_id from workprogramsapp_workprogramchangeindisciplineblockmodule;
    rpd_change_in_rpd_block_module_selection: List[Tuple[int, int]]

    # select id, descipline_block_id from workprogramsapp_disciplineblockmodule;
    rpd_block_module_selection: List[Tuple[int, int]]

    # select id, academic_plan_id from workprogramsapp_disciplineblock;
    rpd_block_selection: List[Tuple[int, int]]

    # select id from workprogramsapp_academicplan;
    academic_plan_selection: List[Tuple[int]]

    # select academic_plan_id, id from workprogramsapp_implementationacademicplan;
    implementation_academic_plan_selection: List[Tuple[int, int]]

    # select implementationacademicplan_id, fieldofstudy_id
    # from workprogramsapp_implementationacademicplan_field_of_study;
    m2m_implementation_academic_plan2field_of_study_selection: List[Tuple[int, int]]

    # select id, number, qualification, title, faculty from workprogramsapp_fieldofstudy_table;
    workprogramsapp_fieldofstudy_selection: List[Tuple[int, str, str, str, Optional[str]]]

    # Внутренние переменные, задаваемые автоматически
    rpd_names: dict = {}
    items_names: dict = {}
    prerequisites: dict = {}
    postrequisites: dict = {}
    rpd_in_field_of_study: Tuple[Tuple[int, int]] = ()
    rpd_change_in_rpd_block_module: Tuple[Tuple[int, int]] = ()
    rpd_block_module: Tuple[Tuple[int, int]] = ()
    rpd_block: Tuple[Tuple[int, int]] = ()
    academic_plan: Tuple[int] = ()
    implementation_academic_plan: Tuple[Tuple[int, int]] = ()
    m2m_implementation_academic_plan2field_of_study: Tuple[Tuple[int, int]] = ()
    workprogramsapp_fieldofstudy: Tuple[Tuple[int, str, str, str, str]] = ()

    graph_data: Tuple[Tuple[int, int, float]] = ()

    @root_validator
    def prepare_data(cls, values):
        for rpd_id, rpd_title in values['rpd_selection']:
            values['rpd_names'][rpd_id] = rpd_title

        for item_id, item_name in values['items_selection']:
            values['items_names'][item_id] = item_name

        for rpd_id, _ in values['rpd_selection']:
            values['prerequisites'][rpd_id] = []
            values['postrequisites'][rpd_id] = []

        for item_id, rpd_id in values['prerequisites_selection']:
            values['prerequisites'][rpd_id].append(item_id)

        for item_id, rpd_id in values['postrequisites_selection']:
            values['postrequisites'][rpd_id].append(item_id)

        values['rpd_in_field_of_study'] = ((id1, id2) for id1, id2 in values['rpd_in_field_of_study_selection'])
        values['rpd_change_in_rpd_block_module'] = ((id1, id2) for id1, id2 in
                                                    values['rpd_change_in_rpd_block_module_selection'])
        values['rpd_block_module'] = ((id1, id2) for id1, id2 in values['rpd_block_module_selection'])
        values['rpd_block'] = ((id1, id2) for id1, id2 in values['rpd_block_selection'])
        values['academic_plan'] = (id1 for id1 in values['academic_plan_selection'])
        values['implementation_academic_plan'] = ((id1, id2) for id1, id2 in
                                                  values['implementation_academic_plan_selection'])
        values['m2m_implementation_academic_plan2field_of_study'] = ((id1, id2) for id1, id2 in values[
            'm2m_implementation_academic_plan2field_of_study_selection'])
        values['workprogramsapp_fieldofstudy'] = ((id1, number, qualification, title, faculty) for
                                                  id1, number, qualification, title, faculty in
                                                  values['workprogramsapp_fieldofstudy_selection'])

        values['graph_data'] = []
        for from_rpd_id, from_rpd_postrequisites in values['postrequisites'].items():
            for to_rpd_id, to_rpd_prerequisites in values['prerequisites'].items():
                if from_rpd_id == to_rpd_id:
                    continue
                weight = count_weight(from_rpd_postrequisites, to_rpd_prerequisites)
                if weight:
                    values['graph_data'].append((from_rpd_id, to_rpd_id, weight))
        values['graph_data'] = tuple(values['graph_data'])

        del values['rpd_selection']
        del values['items_selection']
        del values['prerequisites_selection']
        del values['postrequisites_selection']
        del values['rpd_in_field_of_study_selection']
        del values['rpd_change_in_rpd_block_module_selection']
        del values['rpd_block_module_selection']
        del values['rpd_block_selection']
        del values['academic_plan_selection']
        del values['implementation_academic_plan_selection']
        del values['m2m_implementation_academic_plan2field_of_study_selection']
        del values['workprogramsapp_fieldofstudy_selection']
        collect()

        return values



from typing import Any, Iterator, Dict, Optional, List


def intersected_prerequisites(rpd_index: int, learned_items: list, prerequisites: dict) -> list:
    """
    Общие темы среди пререквизитов РПД и уже освоенных

    :param rpd_index: ID РПД
    :param learned_items: освоенные темы
    :param prerequisites: пререквизиты всех РПД
    :return: список общих тем
    """

    _prerequisites = prerequisites[rpd_index]

    return list(set(_prerequisites).intersection(set(learned_items)))


def intersected_postrequisites(rpd_index: int, preferred_items: list, postrequisites: dict) -> list:
    """
    Общие темы среди постреквизитов РПД и желаемых тем

    :param rpd_index: ID РПД
    :param preferred_items: желаемые темы
    :param postrequisites: постреквизиты всех РПД
    :return: список общих тем
    """

    _postrequisites = postrequisites[rpd_index]

    return list(set(_postrequisites).intersection(set(preferred_items)))


def count_base_coefficient(rpd_index: int, learned_items: list, prerequisites: dict) -> float:
    """
    Функция вычисления коэффициента покрытия знания пререквизитов на основании освоенных тем

    :param rpd_index: ID РПД
    :param learned_items: освоенные темы
    :param prerequisites: пререквизиты всех РПД
    :return: коэффициент покрытия
    """

    _prerequisites = prerequisites[rpd_index]

    return round(len(intersected_prerequisites(rpd_index, learned_items, prerequisites)) / len(_prerequisites),
                 3) if _prerequisites else .0


def count_prefer_coefficient(rpd_index: int, preferred_items: list, postrequisites: dict) -> float:
    """
    Функция вычисления коэффициента покрытия общей желаемости постреквизитов на основании желаемых тем

    :param rpd_index: ID РПД
    :param preferred_items: желаемые темы
    :param postrequisites: постреквизиты всех РПД
    :return: коэффициент покрытия
    """

    _postrequisites = postrequisites[rpd_index]

    return round(len(intersected_postrequisites(rpd_index, preferred_items, postrequisites)) / len(_postrequisites),
                 3) if _postrequisites else .0


def count_weight(from_postrequisites: list, to_prerequisites: list) -> float:
    """
    Функция вычисления веса ребра в графе для алгоритма построения детального маршрута к РПД

    :param from_postrequisites: постреквизиты начальной вершины
    :param to_prerequisites: пререквизиты конечной вершины
    :return: вес ребра
    """

    from_postrequisites = set(from_postrequisites)
    to_prerequisites = set(to_prerequisites)

    intersected_count = len(from_postrequisites.intersection(to_prerequisites))
    united_count = len(from_postrequisites.union(to_prerequisites))

    return round(intersected_count / united_count, 3) if united_count else .0


def get_rpd_id(title: str, rpd_names: Dict[int, str]) -> Optional[int]:
    """
    Функция получения ID РПД по его названию

    :param title: название РПД
    :param rpd_names: хэш-таблица названий РПД
    :return: ID РПД или None
    """

    try:
        return [rpd_id for rpd_id, rpd_title in rpd_names.items() if rpd_title == title][0]
    except IndexError:
        pass


def preferred_rpd_list(preferred_items: list, postrequisites: dict, data, count: int) -> Iterator[Dict]:
    """
    Функция получения списка наиболее желаемых РПД на основании списка желаемых тем без учёта освоенных тем

    :param preferred_items: желаемые темы
    :param postrequisites: постреквизиты всех РПД
    :param data: объект, хранящий все исходные данные из БД и гипер параметры
    :param count: максимальное количество РПД в выдаче
    :return: упорядоченный генератор РПД
    """

    selection = list(map(
        lambda x: (x, count_prefer_coefficient(x, preferred_items, postrequisites)),
        postrequisites.keys(),
    ))
    selection = sorted(selection, key=lambda x: x[1], reverse=True)[: count]

    for rpd, coefficient in selection:
        yield dict(
            rpd_id=rpd,
            rpd_title=data.rpd_names[rpd],
            fields_of_studies=get_fields_of_studies4rpd(rpd, data),
            rate=coefficient,
            common_items=intersected_postrequisites(rpd, preferred_items, postrequisites),
        )


def get_fields_of_studies4rpd(rpd: Any, data, is_id=True) -> List[Dict]:
    """
    Функция получения направлений подготовки для заданной РПД

    :param rpd: ID или название РПД
    :param data: объект, хранящий все исходные данные из БД и гипер параметры
    :param is_id: является ли rpd ID. В обратном случае rpd является названием
    :return: список направлений подготовки
    """

    rpd_id = rpd if is_id else get_rpd_id(rpd, data.rpd_names)

    rpd_change_in_rpd_block_module_id_list = list(
        map(lambda x: x[1], filter(lambda x: x[0] == int(rpd_id), data.rpd_in_field_of_study)))

    rpd_block_module_id_list = list(map(
        lambda x: x[1],
        filter(lambda x: x[0] in rpd_change_in_rpd_block_module_id_list, data.rpd_change_in_rpd_block_module)
    ))

    descipline_block_id_list = list(
        map(lambda x: x[1], filter(lambda x: x[0] in rpd_block_module_id_list, data.rpd_block_module))
    )

    academic_plan_id_list = list(
        map(lambda x: x[1], filter(lambda x: x[0] in descipline_block_id_list, data.rpd_block))
    )

    implementation_academic_plan_id_list = list(
        map(lambda x: x[1], filter(lambda x: x[0] in academic_plan_id_list, data.implementation_academic_plan)))

    field_of_study_id_list = list(map(lambda x: x[1], filter(lambda x: x[0] in implementation_academic_plan_id_list,
                                                             data.m2m_implementation_academic_plan2field_of_study)))

    field_of_study_list = filter(lambda x: x[0] in field_of_study_id_list, data.workprogramsapp_fieldofstudy)

    fos_list = []

    for fos in field_of_study_list:
        fos_list.append(dict(
            field_of_study_name=fos[3],
            field_of_study_code=fos[1],
            faculty=fos[4],
            level=fos[2],
        ))

    return fos_list



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['itmo_routespd', 'itmo_routespd.dijkstra']

package_data = \
{'': ['*']}

install_requires = \
['pydantic==1.8.1']

setup_kwargs = {
    'name': 'itmo-routespd',
    'version': '1.0.0',
    'description': 'Recommendation individual educational routes for ITMO university entrants',
    'long_description': '# ITMO RoutesPD\nRecommendation individual educational routes for ITMO university entrants\n\n*Final thesis by <b>Pivosh Vladislav K3440</b>*\n',
    'author': 'Vladislav Pivosh',
    'author_email': 'pivosh098@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/RUGyron/itmo_routespd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
